/*
 *  Copyright (C) 2003 Intersil Americas Inc.
 *  Copyright 2005 (C) Jean-Baptiste Note
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>

/* ioctl interface -- from kernel module */
#include <islsm_smioctl.h>

static int do_terminal(int upci_fd);

int
main(int argc, char *argv[])
{
	int fd;

	if (argc < 2 || argc > 3) {
		fprintf(stderr,
			"usage: %s islsm_device (for instance /dev/islsm0)\n",
			argv[0]);
		exit(1);
	}

	fd = open(argv[1], O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "Failed opening %s\n", argv[1]);
		exit(1);
	}

	do_terminal(fd);

	close(fd);
	exit(0);
}

static int
do_terminal(int upci_fd)
{
	char buff[1024];
	struct termios otty, ntty;
	void (*old_sigs[32]) (int);
	fd_set mask, rmask;
	register int i;

	printf("You are talking to prism54 device via debug interface\n"
	       "CTRL-] - exit\n"
	       "CTRL-C - reset the card to bootrom\n"
	       "CTRL-U - upload new firmware (via hotplug subsystem)\n"
	       "CTRL-P - send a ping packet to the device\n"
	       "WARNING: these are hardware resets, the driver state is maintained.\n"
	       "Several firmware uploads will lead to error reports in dmesg\n"
	       "which can safely be ignored if you don't use the networking\n"
	       "capabilities of the driver. When loading freemac you may experience\n"
	       "error reports in dmesg, as the code upon firmware upload\n"
	       "in one place still makes the assumption that you are running\n"
	       "an official firmware (it initiates the \"mgmt readback\").\n"
	       "Please disregard them for now\n");
	fflush(stdout);

	/* Set up the keyboard. */
	if (ioctl(0, TCGETS, &otty) < 0) {
		perror("term ioctl TCGETS");
		return (-1);
	}
	ntty = otty;
	ntty.c_lflag &= ~(ECHO | ICANON | ISIG);
	ntty.c_iflag &= ~(ICRNL | IXON | IXOFF);
	ntty.c_oflag &= ~(ONLCR | OPOST);
	if (ioctl(0, TCSETS, &ntty) < 0) {
		perror("term ioctl TCSETS");
		return (-1);
	}

	/* Set up the correct descriptor masks for select(2). */
	FD_ZERO(&mask);
	FD_SET(STDIN_FILENO, &mask);
	FD_SET(upci_fd, &mask);

	/* Disable all signals, saving them for later. */
	for (i = 0; i < 32; i++)
		old_sigs[i] = signal(i + 1, SIG_IGN);

	/* Go into an endless terminal loop. */
	while (1) {
		rmask = mask;
		i = select(32, &rmask, (fd_set *) NULL, (fd_set *) NULL,
			   (struct timeval *)NULL);

		if (i <= 0)
			break;

		if (FD_ISSET(STDIN_FILENO, &rmask)) {
			i = read(STDIN_FILENO, buff, 1);
			if (i > 0) {
				int err = 0;
				if (buff[0] == (']' & 037))
					break;
				else if (buff[0] == 003) {
					printf("\r\nResetting to bootrom...");
					err = ioctl(upci_fd, ISLSM_IOCBOOTROM, 0);
				} else if (buff[0] == ('u' & 037)) {
					printf("\r\nUploading firmware...");
					err = ioctl(upci_fd, ISLSM_IOCLOADFW, 0);
				} else if (buff[0] == ('p' & 037)) {
					printf("\r\nSending ping packet...");
					err = ioctl(upci_fd, ISLSM_IOCSENDPING, 0);
				} else {
					(void)write(upci_fd, buff, i);
					continue;
				}
				if (!err)
					printf("OK\r\n");
				else
					printf("failed\r\n");
			}
		}

		if (FD_ISSET(upci_fd, &rmask)) {
			i = read(upci_fd, buff, 1024);
			if (i > 0)
				(void)write(1, buff, i);
		};
	}

	if (ioctl(0, TCSETS, &otty) < 0) {
		perror("term ioctl TCSETS");
		return (-1);
	}

	/* Restore all signals. */
	for (i = 0; i < 32; i++)
		signal(i + 1, old_sigs[i]);

	printf("\nExiting\n");
	return (0);
}
